#############
# $Id: create_lane_tag_file.pm 13162 2011-05-13 08:58:15Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2010-01-27
# Last Changed On: $Date: 2011-05-13 09:58:15 +0100 (Fri, 13 May 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-27.0/lib/npg_pipeline/analysis/create_lane_tag_file.pm $

package npg_pipeline::analysis::create_lane_tag_file;

use Moose;
use MooseX::InsideOut;
use Carp;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13162 $ =~ /(\d+)/mxs; $r; };
use List::Util qw(max min);

extends q{npg_pipeline::base};
with q{npg_common::roles::run::path_info};
with q{npg_common::roles::run::long_info};
with q{npg_pipeline::roles::business::tag_info};

=head1 NAME

npg_pipeline::analysis::create_lane_tag_file

=head1 VERSION

$LastChangedRevision: 13162 $

=head1 SYNOPSIS

  my $oCreateLaneTagFile = npg_pipeline::analysis::create_lane_tag_file->new({
    id_run => $iIdRun,
    run_folder => $sRunFolder,
    # ... or various paths, something here is required ...
  });

=head1 DESCRIPTION

Object module which knows how to construct lane specific tag files for multiplex runs

=head1 SUBROUTINES/METHODS

=head2 generate

generates lane specific tag_files for multiplex runs

  eval {
    $oCreateLaneTagFile->generate();
  } or do {
    .. your error handling here ...
  };

where a lane is not multiplexed (indexed) then it will generate a 'dummy' tag file with just a single multiple G sequence

=cut

sub generate {
  my ($self) = @_;

  if ( !$self->is_indexed() ) {
    $self->log(q{Run is not multiplexed (indexed)});
    return 1;
  }

  $self->log(q{Run is indexed, obtaining batch});
  my $batch;
  eval {
    $batch  = $self->batch();
  } or do {
    croak q{Unable to obtain batch};
  };

  foreach my $lane ( @{$batch->lanes()} ) {

    my @tag_numbers;

    # only pool lanes will need to be matched against a particular set of tags
    if ( $lane->pool() ) {

      #foreach my $library ( @{$lane->pool()->{entities}} ) {
      #  push @tag_numbers, $library->tag();
      #}

      my $tag_list = $lane->tag_list();

      if( !$tag_list || (scalar @{$tag_list} == 0) ){
        croak 'No tag information available for '. q{Lane } . $lane->position();
      }

      my ($tag_index_list, $tag_seq_list) = $self->_process_tag_list($tag_list);

      if($tag_index_list && $tag_seq_list) {
        #check length
        $tag_seq_list = $self->_check_tag_length($tag_seq_list);
        #check uniqueness
        $self->_check_tag_uniqueness($tag_seq_list);

        $self->log( q{Lane } . $lane->position() . q{ is going to have lane specific tag file created based on sequencescape.});
        $self->_construct_specific_file_expected_sequence($lane, $tag_index_list, $tag_seq_list);
        next;
      }else{
         $self->log( q{Lane } . $lane->position(). ': no expected tag sequence or index given in sequencescape' );
      }

      @tag_numbers = @{$tag_index_list};

      if ($tag_numbers[0]) {
        $self->log( q{Lane } . $lane->position() . q{ is going to have lane specific tag file created.});
        $self->_construct_lane_specific_file( $lane, @tag_numbers );
      } else {
        croak q{This should be a pool, but no tag numbers identified};
      }

    }
    # any other lanes shouldn't show any tags, so create a dummy file, with only G*<index_length>
    # and 100 as the tag number, so that they should all get tagged 0
    else {
      $self->log( q{Lane } . $lane->position() . q{ is going to have dummy tag file created.});
      $self->_construct_dummy_specific_file( $lane->position() );
    }

  }

  return 1;
}

##############
# private methods

# construct the lane specific file
sub _construct_lane_specific_file {
  my ($self, $lane, @tag_numbers) = @_;

  my $position = $lane->position();

  my %tags_wanted;
  foreach my $tag_num (@tag_numbers) {
    $tags_wanted{$tag_num}++;
  }

  my $index_length = $self->index_length();

  # this will hopefully be replaced by Sequencescape providing the oligo sequences as well as the numbers,
  # but until then, look at the expected file matched on index length, and then select from there the ones
  # with the correct tag numbers (if expected fails, then try using entire index read)

  my $tag_file;
  eval {
    $tag_file = $self->get_tag_file({
      index_length => $index_length,
    });
  } or do {
    my $corrected_tag_part_use_bases = $self->corrected_tag_part_use_bases();
    carp qq{Could not find a legitimate tag file to use with index_length $index_length};
    ($index_length) = $corrected_tag_part_use_bases =~ /I(\d+)/ixms;
    carp qq{Attempting to find legitimate tag file with index_length $index_length};
    eval {
      $tag_file = $self->get_tag_file({
        index_length => $index_length,
      });
    } or do {
      croak qq{Unable to obtain a legitimate tag file - $EVAL_ERROR};
    }
  };

  my @wanted_tags;
  open my $fh, q{<}, $tag_file or croak qq{unable to open $tag_file for reading: $EVAL_ERROR};
  while (my $line = <$fh>) {
    my @temp = split /\s+/xms, $line;
    if ($tags_wanted{$temp[1]}) {
      push @wanted_tags, $line;
    }
  }
  close $fh or croak qq{unable to close $tag_file from reading};

  my $lane_specific_tag_file = $self->runfolder_path() . qq{/lane_$position.tag};

  if ( -e $lane_specific_tag_file ) {
    $self->log( qq{$lane_specific_tag_file already exists, not overwriting});
    return 1;
  }

  open $fh, q{>}, $lane_specific_tag_file or croak qq{unable to open $lane_specific_tag_file for writing};
  print {$fh} @wanted_tags or croak qq{unable to print @wanted_tags to $lane_specific_tag_file};
  close $fh or croak qq{unable to close $lane_specific_tag_file from writing};
  return 1;
}

# construct the dummy file for the non-indexed lanes
sub _construct_dummy_specific_file {
  my ($self, $lane) = @_;

  my $index_length = $self->index_length();

  my $lane_specific_tag_file = $self->runfolder_path() . qq{/lane_$lane.tag};

  if ( -e $lane_specific_tag_file ) {
    $self->log( qq{$lane_specific_tag_file already exists, not overwriting});
    return 1;
  }

  # create a lane specific tag file, but leave it empty
  my $return = qx[touch $lane_specific_tag_file];
  if ($CHILD_ERROR) {
    croak qq{unable to touch $lane_specific_tag_file: $return};
  }

  return 1;
}

sub _construct_specific_file_expected_sequence {
  my ($self, $lane, $tag_index_list, $tag_seq_list) = @_;

  my $position = $lane->position();

  if( scalar @{$tag_index_list} != scalar @{$tag_seq_list} ){
    croak "The number of tag indexes is not the same as tag list:@{$tag_index_list}\n@{$tag_seq_list}";
  }
  my $lane_specific_tag_file = $self->runfolder_path() . qq{/lane_$position.tag};

  if ( -e $lane_specific_tag_file ) {
    $self->log( qq{$lane_specific_tag_file already exists, not overwriting});
    return 1;
  }

  open my $fh, q{>}, $lane_specific_tag_file or croak qq{unable to open $lane_specific_tag_file for writing};
  foreach my $array_index ( 0 .. (scalar @{$tag_index_list} -1) ){
     print {$fh} $tag_seq_list->[$array_index].q{ }.$tag_index_list->[$array_index].qq{\n}  or croak q{unable to print};
  }
  close $fh or croak qq{unable to close $lane_specific_tag_file from writing};
  return 1;
}

sub _check_tag_uniqueness {
   my ($self, $tag_seq_list) = @_;

   my %tag_seq_hash = map { $_ => 1 } @{$tag_seq_list};

   if( scalar keys %tag_seq_hash != scalar @{$tag_seq_list} ){

      croak 'The given tags after trimming are not unique';
   }

   return 1;
}

sub _check_tag_length {
   my ($self, $tag_seq_list) = @_;

   my %tag_length = map {length $_ => 1 } @{$tag_seq_list};

   if (scalar keys %tag_length != 1){
      croak "The given tags are different in length: @{$tag_seq_list}";
   }

   my @lengths = keys %tag_length;
   my $tag_len = shift @lengths;
   my @read_cycle_counts = $self->read_cycle_counts();
   my $index_length = $read_cycle_counts[1];
   if( $tag_len > $index_length ){
      $self->log( 'The given tag length are greater than the index length, which will be trimmed ');
      return [ map {substr $_, 0,$index_length} @{$tag_seq_list} ];
   }

   return $tag_seq_list;
}

sub _process_tag_list {
  my ($self, $tag_list) = @_;

  my @tag_index_list = map {$_->index()} @{$tag_list};
  my @tag_seq_list  = map {$_->expected_sequence()} @{$tag_list};

  foreach my $tag_index (@tag_index_list){
    if(!$tag_index){
      $self->log('The tag index are not available from sequencescape');
      return (undef, undef);
    }
  }

  foreach my $tag_seq (@tag_seq_list){
    if(!$tag_seq){
      $self->log('The tag sequence are not available from sequencescape');
      return (\@tag_index_list, undef);
    }
  }

  my $trimmed_tag_seq_list = $self->_trim_tag_common_suffix(\@tag_seq_list);

  return (\@tag_index_list, $trimmed_tag_seq_list);
}

#trim common suffix of a list of tag sequences
#croak if the tag sequence are different in length or they are all the same
sub _trim_tag_common_suffix {
  my ($self, $tag_seq_list) = @_;

  if(scalar @{$tag_seq_list} == 1){
    carp 'Only one tag found for this lane';
    return $tag_seq_list;
  }

  my %tag_length = map {length $_ => 1 } @{$tag_seq_list};

  if (scalar keys %tag_length != 1){
    croak "The given tags are different in length: @{$tag_seq_list}";
  }

  my $tag_common_suffix_length = $self->_tag_common_suffix_length($tag_seq_list);

  if($tag_common_suffix_length == 0){
    carp 'No need to trim given tags';
    return $tag_seq_list;
  }

  if($tag_length{$tag_common_suffix_length}){
    croak "All tags are the same @{$tag_seq_list}";
  }

  my @trimmed_tag_seq_list;

  foreach my $tag (@{$tag_seq_list}){
    push @trimmed_tag_seq_list, substr $tag, 0, - $tag_common_suffix_length;
  }

  return \@trimmed_tag_seq_list;
}

#find the common suffix of length for a list tag sequences
sub _tag_common_suffix_length {
   my ($self, $tag_seq_list) = @_;

   my $common_suffix_length = 0;

   my %tag_length = map {length $_ => 1 } @{$tag_seq_list};

   my $shotest_tag_length = min keys %tag_length;

   while($common_suffix_length < $shotest_tag_length){

      $common_suffix_length++;
      my $previous_suffix = substr $tag_seq_list->[0], - $common_suffix_length;

      foreach my $tag (@{$tag_seq_list}){

        my $current_suffix = substr $tag, - $common_suffix_length;
        if($current_suffix ne $previous_suffix){
          return $common_suffix_length - 1;
        }
      }
   }

   return $common_suffix_length;
}

no Moose;
__PACKAGE__->meta->make_immutable;

1;

__END__


=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Carp

=item English -no_match_vars

=item Readonly

=item Moose

=item npg_pipeline::base

=item npg_common::roles::run::path_info

=item npg_common::roles::run::long_info

=item npg_pipeline::roles::business::tag_info

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
