#############
# $Id: to_sra.pm 13162 2011-05-13 08:58:15Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2009-09-01
# Last Changed On: $Date: 2011-05-13 09:58:15 +0100 (Fri, 13 May 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-27.0/lib/npg_pipeline/archive/file/to_sra.pm $

package npg_pipeline::archive::file::to_sra;
use Moose;
use MooseX::InsideOut;
use Carp;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 13162 $ =~ /(\d+)/mxs; $r; };
use Digest::MD5;
use Perl6::Slurp;

extends qw{npg_pipeline::archive};

Readonly::Scalar our $ARCHIVE_SCRIPT      => q{npg_pipeline_archive_to_sra};

Readonly::Scalar our $SHIFT_ERROR_CODE    => 8;
Readonly::Scalar our $LAST_INDEX          => -1;

sub submit_to_lsf {
  my ($self, $arg_refs) = @_;
  my $job_sub = $self->_generate_bsub_command($arg_refs);
  my $job_id = $self->submit_bsub_command($job_sub);;
  return ($job_id);
}

sub archive {
  my ($self) = @_;

  my $run = $self->run();

  # Environment hacking for Oracle
  my $oracle_home = defined ($ENV{ORACLE_HOME}) ? $ENV{ORACLE_HOME} : $self->source_files_conf->{oracle_home};
  my @ora_libs = ( split ( /:/mxs, $ENV{LD_LIBRARY_PATH} || q[] ), $self->_oracle_libs() );
  local $ENV{ORACLE_HOME}     = $oracle_home;
  local $ENV{LD_LIBRARY_PATH} = join q[:], @ora_libs;

  # get array of run_lane objects
  my $run_lanes   = $run->run_lanes;
  if(!scalar @{$run_lanes}) {
    $self->log(q[_run_archive_to_sra: Nothing to do]);
    return 1;
  }

  my $archive_dir = $self->archive_path();

  my $id_run    = $run->id_run();
  $self->log($id_run);
  my $run_name  = $run->name();
  $self->log($run_name);

  for my $run_lane (@{$run_lanes}) {
    $self->_upload_run_lane_data({
      run_lane => $run_lane,
      id_run   => $id_run,
      run_name => $run_name,
      run      => $run,
      instrument_model => $self->instrument_type(),
    });
  }

  return 1;
}

sub _oracle_libs {
  my ( $self ) = @_;
  my $oracle_home = $self->source_files_conf->{oracle_home};
  return (qq[$oracle_home/lib], qq[$oracle_home]);
}

sub _upload_run_lane_data { ## no critic (Subroutines::ProhibitExcessComplexity)
  my ($self, $arg_refs) = @_;

  my $run_lane  = $arg_refs->{run_lane};
  my $id_run    = $arg_refs->{id_run};
  my $run_name  = $arg_refs->{run_name};
  my $run       = $arg_refs->{run};
  my $instrument_is_hiseq = 0;
  if ( $arg_refs->{instrument_model} eq q{HiSeq} ) {
    $instrument_is_hiseq = 1;
  }

  my $lane_pos = $run_lane->position();

  $self->log(sprintf q[SRA archival for %d / %d], $id_run, $lane_pos);

  # See if this lane has already got files loaded
  my ( $existing_count, $existing_files ) = $self->_check_already_uploaded($id_run, $lane_pos);
  if ( $existing_count ) {
    $self->log( q{For } . $id_run . q{_} . $lane_pos . q{ : } . $existing_count . q{ files already present.} );
    my $msg = qq{Already present files\n};
    foreach my $file ( sort keys %{ $existing_files } ) {
      $msg .= qq{\t$file\n};
    }
    $self->log( $msg );
  }

  # Find files available for SRA archival. This logic supports the
  # presence of more than one file of a given type -- e.g., separate
  # fastq files for the two reads of a paired-end run

  $self->log(q{Looking for files in } . $self->archive_path() . q{/} . $id_run . q{_} . $lane_pos. q{*});

  my @files_to_load;
  my @non_consented_files_to_load;
  my @hiseq_group_files_to_load;
  my $consented_files;
  my $xref_sources = {run => $run};
  my @xrefs_to_load;

  my $library = $run_lane->library;
  my $pool = $run_lane->pool();
  my $control  = $run_lane->control();

  if ( $library ) {
    $self->log( q{library} );
  }
  if ( $pool ) {
    $self->log( q{pool} );
  }
  if ( $control ) {
    $self->log( q{control} );
  }
  my $control_is_phix;

  if ( $library ) {

    $xref_sources->{'library'} = $library;

    my $project = $library->study;
    if($project) {
      $xref_sources->{'project'} = $project;
    } else {
      $self->log(sprintf q[Could not determine project from library %d. No xref generated.], $library->id());
    }


  } else {
    my $msg = q{};
    if ( $pool ) {
      $msg .= q{This lane is a pool. };
    } elsif ( $control ) {
      if ( $control->name() =~ /phix/ixms ) {
        $control_is_phix++;
        $msg .= q{This lane is } . $control->name() . q{. };
      } else {
        $msg .= q{This lane thinks it is a control, but does not appear to be phix. };
      }
    }
    if ( ! $msg ) {
      $msg .= q{This is not a library, but I don't know what it think it is. };
    }
    $self->log( $msg . q{No xref generated.} );
  }

  my @types_to_load = ( q{fastqcheck} );
  if ( ! $control_is_phix ) {
    unshift @types_to_load, ( qw{ fastq } );
  }

  for my $type ( @types_to_load ) {

    opendir my $dh, $self->archive_path();
    my @dir_files = readdir $dh;
    closedir $dh;
    my @fn;
    foreach my $dir_file ( @dir_files ) {
      if ( $dir_file =~ m/${id_run}_s?_?${lane_pos}/xms && $dir_file =~ /[.]${type}\z/xms && $dir_file !~ /orig/ixms ) {
        push @fn, $self->archive_path() . q{/} . $dir_file;
      }
    }
    if (@fn == 0) {
      $self->log(qq[ +- No $type for $id_run / $lane_pos]);
      next;
    }

    if ( $type eq q{fastq} ) {
      foreach my $filename (@fn) {
        if ( $filename =~ /human/xms ) {
          $consented_files++;
        }
      }
    }

    foreach my $filename (@fn) {
      my @temp = split m{/}xms, $filename;

      if ( ! $existing_files->{ $temp[$LAST_INDEX] } ) {
        my ($dtfn) = $filename =~ /([\w\/\,\.\-]+)/xms;
        my $load = 1;
        my $href_file_to_load = {};
        if ( $consented_files && $type eq q{fastq} ) {
          if ($filename !~ /human/xms) {
            undef$load;
            next;
          } else {
            if ($filename =~ /_non/xms) {
              if ( $instrument_is_hiseq ) { ## no critic (ControlStructures::ProhibitDeepNests)
                push @hiseq_group_files_to_load, {type => $type, name => $dtfn};
              } else {
                push @files_to_load, {type => $type, name => $dtfn};
              }
            } else {
              push @non_consented_files_to_load, {type => $type, name => $dtfn};
            }
          }
        } else {
          if ( $instrument_is_hiseq && $type eq q{fastq} ) {
            push @hiseq_group_files_to_load, {type => $type, name => $dtfn};
          } else {
            push @files_to_load, {type => $type, name => $dtfn};
          }
        }
        if ($load) {
          $self->log(qq[ +- Planning to load $type:$dtfn]);
        }
      }
    }

  }

  my @all_files_to_load = ( @files_to_load, @non_consented_files_to_load, @hiseq_group_files_to_load );

  if ( ! scalar @all_files_to_load) {
    $self->log( q[No files available or required to load] );
    return 1;     # next lane
  }

  $self->log( q[Loading ] . scalar @all_files_to_load . q[ files]);

  # Build xrefs for SRA archival.

  for my $xref_source_type (keys %{$xref_sources}) {

    my $xs    = $xref_sources->{$xref_source_type};
    my $pk    = $xs->primary_key;
    my $id    = $xs->$pk();
    my $name  = $xs->name();

    $self->log(qq[Planning to xref $xref_source_type/$id/$name]);

    # Backslashify dodgy characters.  This is mainly needed to deal with
    # slashes that may have been put into project or sample names.
    $name =~ s{([^A-Za-z0-9_-])}{\\$1}gmxs;
    $name =~ s{\/}{\\\/}gxms;
    $name =~ s{\/}{\\\/}gxms;

    push @xrefs_to_load, {
                          type => $xref_source_type,
                          id   => $id,
                          name => $name,
                         };
  }

  # construct commandline for LSF

  my @files = map { ('-f', sprintf '%s:%s', $_->{type}, $_->{name} ) }
              @files_to_load;

  my @xrefs = map { ('-x', sprintf '%s/id=%d/name=%s',
              $_->{type}, $_->{id}, $_->{name} ) }
              @xrefs_to_load;

  my $collection_name = sprintf '%d_s_%d', $id_run, $lane_pos ;

  my $cmd  = $self->internal_script_bin() .  q{/} . $self->external_script_names_conf()->{mpsa_upload_exe};
  $cmd    .= qq{ -c $collection_name };
  $cmd    .= join q{ }, @xrefs;
  $cmd    .= q{ };
  $cmd    .= join q{ }, @files;

  $self->log(qq{sra command: $cmd});

  my $rc = qx/$cmd/;
  $self->log(qq{return code:$CHILD_ERROR});
  if ($CHILD_ERROR != 0) {
    $self->log(qq{Upload unsuccessful - Error code: $CHILD_ERROR});
    croak q{Error uploading.};
  }
  $self->log(q{Upload successful});

  if ( $consented_files ) {
    my @non_consented_files = map { ('-f', sprintf '%s:%s', $_->{type}, $_->{name} ) }
                              @non_consented_files_to_load;

    $cmd  = $self->internal_script_bin() .  q{/} . $self->external_script_names_conf()->{mpsa_upload_exe};
    $cmd .= qq{ -c $collection_name};
    $cmd .=  q{ -g } . $self->general_values_conf()->{unconsented_sra_group} . q{ };
    $cmd .= join q{ }, @non_consented_files;

    $self->log(qq{Unconsented sra command: $cmd});

    $rc = qx/$cmd/;
    if ($CHILD_ERROR != 0) {
      $self->log(qq{Upload unsuccessful - Error code: $CHILD_ERROR});
      croak q{Error uploading.};
    }
    $self->log(q{Upload successful});

    $self->log(qq{Unconsented upload output: $rc});
  }

  if ( $instrument_is_hiseq ) {
    my @hiseq_group_files = map { ('-f', sprintf '%s:%s', $_->{type}, $_->{name} ) }
                              @hiseq_group_files_to_load;

    $cmd  = $self->internal_script_bin() .  q{/} . $self->external_script_names_conf()->{mpsa_upload_exe};
    $cmd .= qq{ -c $collection_name};
    $cmd .=  q{ -g } . $self->general_values_conf()->{hiseq_sra_group} . q{ };
    $cmd .= join q{ }, @hiseq_group_files;

    $self->log(qq{HiSeq loading sra command: $cmd});

    $rc = qx/$cmd/;
    if ($CHILD_ERROR != 0) {
      $self->log(qq{Upload unsuccessful - Error code: $CHILD_ERROR});
      croak q{Error uploading.};
    }
    $self->log(q{Upload successful});

    $self->log(qq{HiSeq file upload output: $rc});
  }

  return $self->_check_md5sum( {
    id_run => $id_run,
    lane => $lane_pos,
    control_is_phix => $control_is_phix,
    instrument_is_hiseq => $instrument_is_hiseq,
  } );
}

# check the md5sums of the files uploaded in the fuse directory against the md5sum file we have
sub _check_md5sum {
  my ( $self, $arg_refs ) = @_;
  my $id_run              = $arg_refs->{id_run};
  my $lane                = $arg_refs->{lane};
  my $control_is_phix     = $arg_refs->{control_is_phix};
  my $instrument_is_hiseq = $arg_refs->{instrument_is_hiseq};

  $self->log( qq{Checking md5sums for $id_run, $lane} );

  # if the lane is control and phix, then don't bother checking the md5sums just for fastqcheck files
  if ( $control_is_phix ) {
    $self->log( q{This is a phix control lane, not checking md5sums for only fastqcheck} );
    return 1;
  }

  my $md5_hash = $self->_lane_files_md5_vals()->{$lane};

  if (!$md5_hash) {
    $self->log(qq{No md5 values for any files relating to $id_run, $lane - assuming ok});
    return 1;
  }

  my @incorrect_files;

  foreach my $file (sort keys %{$md5_hash}) {

    # we won't have loaded any srf's, but runs analysed before release 25.0 will have generated md5s
    if ( $file =~ /srf/xms ) {
      next;
    }
#    my $file_path = $self->fuse_md5_archive($id_run) . q{/} . $file;

    # human split md5 values can only be obtained if we use the download command
    my $cmd = $self->internal_script_bin() . q{/} . $self->external_script_names_conf()->{mpsa_download_exe} . qq{ -m -f $file};
    my $content = qx[$cmd];

    # If an error occurs here, add to the incorrect files and move to next;
    if ($CHILD_ERROR != 0) {
      my $msg = qq{Unable to obtain file $file - Error code: $CHILD_ERROR, output: $content};
      $self->log( $msg );
      push @incorrect_files, $msg;
      next;
    }

    $content ||= q{};

    $self->log( qq{Content of $file: $content} );

    my ($hex, @temp) = split /\s+/xms, $content;

    $hex ||= q{};

    if ($hex ne $md5_hash->{$file}) {
      my $msg = qq{$file has incorrect md5:\n\tFilesystem:\t$md5_hash->{$file}\n\tFuse:\t\t$hex\n};
      $self->log($msg);
      push @incorrect_files, $msg;
    } else {
      $self->log(qq{$file has correct md5});
    }
  }

  if (scalar @incorrect_files) {
    croak @incorrect_files;
  }
  return 1;
}

has q{_lane_files_md5_vals} => (isa => q{HashRef}, is => q{ro}, lazy_build => 1, writer => q{_lane_files_md5_vals_for_test});

sub _build__lane_files_md5_vals {
  my ($self) = @_;

  my $hash = {};

  my $id_run = $self->id_run();
  foreach my $run_lane (@{$self->run->run_lanes()}) {
    my $file = $self->archive_path() . q{/} . $id_run . q{_} . $run_lane->position() . q{.md5};
    if (! -e $file) {
      $self->log( qq{$file does not exist - cannot check md5s for this lane} );
      next;
    }
    my @md5s = slurp ( $file );
    if (! scalar @md5s) {
      $self->log( qq{no md5s where calculated in $file} );
    }
    foreach my $md5_line ( @md5s ) {
      next if ($md5_line =~ /gcfreq/xms);
      my @temp = split /\s+/xms, $md5_line;
      my @file_path = split m{/}xms, $temp[1];
      my $file_name = $file_path[$LAST_INDEX];
      $hash->{$run_lane->position()}->{$file_name} = $temp[0];
    }

  }
  return $hash;
}

# check to see if a lane has already been uploaded
# would like this to check which files have
sub _check_already_uploaded {
  my ($self, $id_run, $lane) = @_;

  my $count = 0;

  my $cmd = $self->internal_script_bin() .  q{/} . $self->external_script_names_conf()->{mpsa_download_exe};
  $cmd   .= q{ -l};
  $cmd   .= q{ -s};
  $cmd   .= q{ 'c:M_COLLECTION_NAME = } . $id_run . q{_s_} . $lane . q{'};

  $self->log("check if uploaded running: $cmd");

  my $existing_files = {};

  eval {

    my $uploaded_list = qx/$cmd/;
    if ( $CHILD_ERROR ) {
      $self->log(qq{Error running: $cmd});
      croak(qq{Error running $cmd\nReturned error code: $CHILD_ERROR, quitting.\n});
    }

    if ( $uploaded_list ) {
      chomp$uploaded_list;
      my @files = split m{\n}xms, $uploaded_list;
      $count = scalar @files;
      foreach my $file ( @files ) {
        $existing_files->{$file}++;
      }
    }
    1;
  } or do {
    $self->log($EVAL_ERROR);
  };

  return ( $count, $existing_files );
}

sub _generate_bsub_command {
  my ($self, $arg_refs) = @_;

  my $required_job_completion = $arg_refs->{required_job_completion};
  my $timestamp = $self->timestamp();

  my $job_name = $ARCHIVE_SCRIPT . q{_} . $self->id_run() . q{_} . $timestamp;

  my $command_line_options = $self->attributes_as_command_options( {
    excluded_attributes => [ qw{ log_file_name log_file_path log_file timestamp array_cpu_limit } ]
  });

  my $location_of_logs = $self->make_log_dir( $self->recalibrated_path() );
  my $bsub_command = q{bsub -q } . $self->lsf_queue() . qq{ $required_job_completion -J $job_name };
  $bsub_command .=  ( $self->fs_resource_string( {
    counter_slots_per_job => 1,
  } ) ) . q{ };
  $bsub_command .=  q{-o } . $location_of_logs . qq{/$job_name.out };
  $bsub_command .=  q{'} . $self->internal_script_bin() . q{/} . $ARCHIVE_SCRIPT . q{ };
  $bsub_command .= qq{$command_line_options'};
  if ($self->verbose()) {
    $self->log($bsub_command);
  }
  return $bsub_command;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 NAME

npg_pipeline::archive::file::to_sra

=head1 VERSION

$LastChangedRevision: 13162 $

=head1 SYNOPSIS

  my $fsa = npg_pipeline::archive::file::to_sra->new({
    run_folder => <run_folder>,
    timestamp => $sTimeStamp,
  });

=head1 DESCRIPTION

=head1 SUBROUTINES/METHODS

=head2 submit_to_lsf

handler for submitting to LSF the archival to MPSA the fastq, fastqcheck and gcfreq files
returns an array of lsf job ids

  my @job_ids = $fsa->submit_to_lsf({
    required_job_completion => <string of lsf job dependencies>,
  });

=head2 archive

wrapper method which goes through each run_lane on the run, and decides whether it has already been archived, or needs archiving,
and then runs the loading scripts

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::InsideOut

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
