# $Id: 60-reorder_fastq.t 13162 2011-05-13 08:58:15Z ajb $
use strict;
use warnings;
use Carp;
use English qw{-no_match_vars};
use Test::More;
use Test::Exception;
use lib qw{t};
use t::util;
use List::MoreUtils qw(uniq);

if (!$ENV{REORDER_FASTQ}) {
  my $msg = 'Reorder fastq test not run (slow tests).  Set $ENV{REORDER_FASTQ} to a true value to run.';
  plan( skip_all => $msg );
}

plan( tests => 13 );

use_ok('npg_pipeline::reorder_fastq');

my $util = t::util->new({});
my $analysis_runfolder_path = $util->analysis_runfolder_path();

sub create_archive_dir {
  my $tmp_staging = $util->temp_directory() . q{/staging};
  `rm -rf $tmp_staging`;
  `mkdir -p $analysis_runfolder_path/Latest_summary/archive`;
  return 1;
}

sub setup_fastq_files {
  my ($no_tags) = @_;
  create_archive_dir();
  `cp t/data/test_reordering_fastqs/* $analysis_runfolder_path/Latest_summary/archive/`;
  if ($no_tags) {
    `rm -f $analysis_runfolder_path/Latest_summary/archive/*t.fastq`;
  }
  return 1;
}

sub test_correct_order {
  my ($dir, $position, $orig_dir) = @_;

  my @fd = glob "$dir/*_".$position."_*.fastq";
  if ($orig_dir) {
    my @files = glob "$orig_dir/*_".$position."_1.fastq";
    my $file = shift @files;
    unshift @fd, $file;
  }

  my $tl = badidorder(@fd);
  if ($tl) {
    return "$dir $position $tl";
  }
  return;
}

sub badidorder {
  my @fd = @_;

  return if (@fd<2);

  my @fh = map {  open(my $fh,q(<),$_);
                  $fh
               }  @fd;
  my $li = 0;
  while (1) {
    my @l = map { my $l=<$_>; } @fh;
    last unless(@l and defined($l[0]));
    @l = map { m|\A(\S+)/| } @l;
    if (not ($li++%4)) {
      if (1 < scalar(uniq @l)) {
        return $li
      }
    }; 
  }
}; 
{
  setup_fastq_files(q{no tags});
  my $reorder_object;
  my $path = $util->temp_directory() . q{/staging/IL2/analysis/123456_IL2_1234/Latest_summary/archive};
  throws_ok { $reorder_object = npg_pipeline::reorder_fastq->new({
    fastq2 => q{test_1_2.fastq},
    path => $path,
  }); } qr{Attribute[ ][(]fastq1[)][ ]is[ ]required}, q{no creation without fastq1};
  throws_ok { $reorder_object = npg_pipeline::reorder_fastq->new({
    fastq1 => q{test_1_1.fastq},
    path => $path,
  }); } qr{Attribute[ ][(]fastq2[)][ ]is[ ]required}, q{no creation without fastq2};
  throws_ok { $reorder_object = npg_pipeline::reorder_fastq->new({
    fastq2 => q{test_1_2.fastq},
    fastq1 => q{test_1_1.fastq},
  }); } qr{Attribute[ ][(]path[)][ ]is[ ]required}, q{no creation without path};
  lives_ok { $reorder_object = npg_pipeline::reorder_fastq->new({
    fastq2 => q{test_1_2.fastq},
    fastq1 => q{test_1_1.fastq},
    path => $path,
  }); } q{successfuly created};
  lives_ok { $reorder_object->run(); } q{no croak on running only R1 and R2};
  my $result = test_correct_order($path, 1);
  ok(!$result, q{generated file is as expected});
}

{
  setup_fastq_files();
  my $path = $util->temp_directory() . q{/staging/IL2/analysis/123456_IL2_1234/Latest_summary/archive};
  my $reorder_object = npg_pipeline::reorder_fastq->new({
    fastqt => q{test_1_t.fastq},
    fastq2 => q{test_1_2.fastq},
    fastq1 => q{test_1_1.fastq},
    path => $path,
  });
  lives_ok { $reorder_object->run(); } q{no croak on running with multiplex};
  ok(!test_correct_order($path, 1), q{generated files are as expected with multiplex});
}

{
  create_archive_dir();
  my $path = q{t/data/test_reordering_fastqs};
  my $new_path = $util->temp_directory() . q{/staging/IL2/analysis/123456_IL2_1234/Latest_summary/archive};
  my $reorder_object = npg_pipeline::reorder_fastq->new({
    fastqt    => q{test_1_t.fastq},
    fastq2    => q{test_1_2.fastq},
    fastq1    => q{test_1_1.fastq},
    path      => $path,
    new_path  => $new_path,
  });
  lives_ok { $reorder_object->run(); } q{no croak on running};
  ok(!test_correct_order($new_path, 1, $path), q{generated files, where output dir is different from input dir are as expected});

  create_archive_dir();
  lives_ok { qx[bin/npg_pipeline_reorder_fastq --fastq1=test_1_1.fastq --fastq2=test_1_2.fastq --path=$path --new_path=$new_path]; } q{npg_pipeline_reorder_fastq runs};
  ok(!$CHILD_ERROR, qq{Error code ok : $CHILD_ERROR});
}
1;